from __future__ import annotations

import typing

from rest_framework import status
from rest_framework.mixins import CreateModelMixin as BaseCreateModelMixin
from rest_framework.mixins import DestroyModelMixin as BaseDestroyModelMixin
from rest_framework.mixins import ListModelMixin as BaseListModelMixin
from rest_framework.mixins import RetrieveModelMixin as BaseRetrieveModelMixin
from rest_framework.mixins import UpdateModelMixin as BaseUpdateModelMixin
from rest_framework.request import Request
from rest_framework.response import Response

from restdoctor.rest_framework.negotiations import APIVersionContentNegotiation
from restdoctor.rest_framework.pagination import PageNumberPagination
from restdoctor.rest_framework.response import ResponseWithMeta
from restdoctor.rest_framework.serializers import EmptySerializer

if typing.TYPE_CHECKING:
    from django.db.models import QuerySet
    from rest_framework.pagination import BasePagination
    from rest_framework.serializers import BaseSerializer


class NegotiatedMixin:
    content_negotiation_class = APIVersionContentNegotiation


class CreateModelMixin(BaseCreateModelMixin):
    def create(self, request: Request, *args: typing.Any, **kwargs: typing.Any) -> Response:
        request_serializer_class = self.get_request_serializer_class()
        serializer = self.get_serializer_instance(
            request_serializer_class, stage='request', data=request.data
        )
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)

        response_serializer_class = self.get_response_serializer_class()
        if response_serializer_class != request_serializer_class:
            serializer = self.get_serializer_instance(
                response_serializer_class, serializer.instance, stage='response'
            )
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class ListModelMixin(BaseListModelMixin):
    pagination_class: typing.Optional[BasePagination] = PageNumberPagination

    def get_serializer(self, *args: typing.Any, **kwargs: typing.Any) -> BaseSerializer:
        return self.get_response_serializer(*args, **kwargs)

    def list(  # noqa: A003
        self, request: Request, *args: typing.Any, **kwargs: typing.Any
    ) -> Response:
        request_serializer = self.get_request_serializer(
            data=request.query_params, use_default=False
        )
        request_serializer.is_valid(raise_exception=True)
        queryset = self.get_collection(request_serializer)
        meta = self.get_meta_serializer_data()
        page = self.paginate_queryset(queryset)
        if page is not None:
            self.perform_list(page)
            serializer = self.get_serializer(page, many=True)
            response = self.get_paginated_response(serializer.data)
            response.meta.update(meta)
            return response

        self.perform_list(queryset)
        serializer = self.get_serializer(queryset, many=True)
        return ResponseWithMeta(data=serializer.data, meta=meta)

    def get_collection(
        self, request_serializer: BaseSerializer
    ) -> typing.Union[typing.List, QuerySet]:
        return self.filter_queryset(self.get_queryset())

    def perform_list(self, data: typing.Union[typing.List, QuerySet]) -> None:
        pass

    def get_meta_data(self) -> typing.Dict[str, typing.Any]:
        return {}

    def get_meta_serializer_data(self) -> typing.Dict[str, typing.Any]:
        if issubclass(self.get_meta_serializer_class(), EmptySerializer):
            return {}
        serializer = self.get_meta_serializer(self.get_meta_data())
        return serializer.data


class RetrieveModelMixin(BaseRetrieveModelMixin):
    def get_serializer(self, *args: typing.Any, **kwargs: typing.Any) -> BaseSerializer:
        return self.get_response_serializer(*args, **kwargs)


class UpdateModelMixin(BaseUpdateModelMixin):
    def update(self, request: Request, *args: typing.Any, **kwargs: typing.Any) -> Response:
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        request_serializer_class = self.get_request_serializer_class()
        serializer = self.get_serializer_instance(
            request_serializer_class, instance, stage='request', data=request.data, partial=partial
        )
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(serializer.instance, '_prefetched_objects_cache', None):
            serializer.instance._prefetched_objects_cache = {}

        response_serializer_class = self.get_response_serializer_class()
        if response_serializer_class != request_serializer_class:
            serializer = self.get_serializer_instance(
                response_serializer_class, serializer.instance, stage='response'
            )
        return Response(serializer.data)


class DestroyModelMixin(BaseDestroyModelMixin):
    pass



from __future__ import annotations

import typing


def is_union_type(tp: typing.Type) -> bool:
    return tp is typing.Union or getattr(tp, '__origin__', None) is typing.Union


def is_optional_type(tp: typing.Type) -> bool:
    if tp is type(None):  # noqa: E721
        return True
    elif is_union_type(tp):
        return any(is_optional_type(tt) for tt in getattr(tp, '__args__', []))
    else:
        return False


def is_list_type(tp: typing.Type) -> bool:
    if tp is list or getattr(tp, '__origin__', None) is list:
        return True
    if is_union_type(tp):
        if (
            len(tp.__args__) == 2
            and is_list_type(tp.__args__[0])
            and is_optional_type(tp.__args__[1])
        ):
            return True
    return False



from __future__ import annotations

import pytest

from restdoctor.rest_framework.schema import RefsSchemaGenerator


@pytest.mark.parametrize(
    'api_format, expected_api_format',
    [
        (('full', 'compact', 'compact-verbose'), ('full', 'compact', 'compact-verbose')),
        (('full', 'compact:{12,32}'), ('full', 'compact:12', 'compact:32')),
        (('compact:{12}',), ('compact:12',)),
        (('test:{1,2,3,4}',), ('test:1', 'test:2', 'test:3', 'test:4')),
        (('full', 'compact:12'), ('full', 'compact:12')),
    ],
)
def test_schema_with_tags_success_case(settings, api_format, expected_api_format):
    settings.API_FORMATS = api_format

    refs_schema = RefsSchemaGenerator()

    assert refs_schema.api_formats == expected_api_format



from __future__ import annotations

import pytest

from restdoctor.utils.serializers import get_serializer_class_from_map
from tests.test_unit.test_serializers.test_serializer_class_map.stubs import (
    SerializerA,
    SerializerB,
    SerializerC,
    SerializerD,
    SerializerE,
    SerializerF,
    SerializerG,
    serializer_class_map_no_default,
    serializer_class_map_with_default,
    serializer_class_map_with_format,
)


@pytest.mark.parametrize(
    'serializer_class_map,action,stage,default,expected',
    (
        (serializer_class_map_with_default, 'retrieve', 'request', SerializerD, SerializerB),
        (serializer_class_map_with_default, 'retrieve', 'response', SerializerD, SerializerA),
        (serializer_class_map_with_default, 'list', 'response', SerializerD, SerializerB),
        (serializer_class_map_no_default, 'retrieve', 'request', SerializerD, SerializerB),
        (serializer_class_map_no_default, 'retrieve', 'response', SerializerD, SerializerD),
        (serializer_class_map_no_default, 'list', 'response', SerializerD, SerializerB),
    ),
)
def test_get_serializer_class_from_map(serializer_class_map, action, stage, default, expected):
    serializer_class = get_serializer_class_from_map(action, stage, serializer_class_map, default)

    assert serializer_class == expected


@pytest.mark.parametrize(
    'serializer_class_map,action,stage,api_format,default,expected',
    (
        (serializer_class_map_with_format, 'retrieve', 'request', None, SerializerD, SerializerB),
        (serializer_class_map_with_format, 'retrieve', 'response', None, SerializerD, SerializerC),
        (
            serializer_class_map_with_format,
            'retrieve',
            'response',
            'full',
            SerializerD,
            SerializerB,
        ),
        (
            serializer_class_map_with_format,
            'list',
            'response',
            'with_related',
            SerializerD,
            SerializerA,
        ),
        (
            serializer_class_map_with_format,
            'create',
            'response',
            'version:12',
            SerializerD,
            SerializerG,
        ),
        (
            serializer_class_map_with_format,
            'create',
            'response',
            'version:1',
            SerializerD,
            SerializerA,
        ),
        (serializer_class_map_with_format, 'list', 'request', None, SerializerD, SerializerB),
        (serializer_class_map_with_format, 'list', 'response', 'full', SerializerD, SerializerE),
        (
            serializer_class_map_with_format,
            'list',
            'response',
            'version:2',
            SerializerD,
            SerializerG,
        ),
        (
            serializer_class_map_with_format,
            'list',
            'response',
            'version:12',
            SerializerD,
            SerializerF,
        ),
        (
            serializer_class_map_with_format,
            'list',
            'response',
            'version:15',
            SerializerD,
            SerializerF,
        ),
        (
            serializer_class_map_with_format,
            'list',
            'response',
            'version:33',
            SerializerD,
            SerializerF,
        ),
        (
            serializer_class_map_with_format,
            'list',
            'response',
            'test:name',
            SerializerD,
            SerializerG,
        ),
    ),
)
def test_get_serializer_class_from_map_with_format(
    serializer_class_map, action, stage, api_format, default, expected, settings
):
    settings.API_DEFAULT_FORMAT = 'compact'
    settings.API_FORMATS = ('full', 'compact', 'version:{2,12,33,15}', 'test:name')

    serializer_class = get_serializer_class_from_map(
        action, stage, serializer_class_map, default, api_format=api_format
    )

    assert serializer_class == expected



import pytest

from tests.test_unit.test_serializers.test_meta_inheritance.stubs import (
    ListViewSetWithMetaSerializer,
    ListViewSetWithRequestSerializer,
)


@pytest.mark.parametrize(
    'viewset,meta_data,result',
    (
            (ListViewSetWithMetaSerializer, {'char_field': 'test'}, {'char_field': 'test'}),
            (ListViewSetWithMetaSerializer, {'char_field': 'test', 'data': {}}, {'char_field': 'test'}),
            (ListViewSetWithMetaSerializer, {'char_field': 'test', 'data': []}, {'char_field': 'test'}),
            (ListViewSetWithRequestSerializer, {'char_field': 'test'}, {}),
            (ListViewSetWithRequestSerializer, {'char_field': 'test', 'data': []}, {}),
            (ListViewSetWithRequestSerializer, {'char_field': 'test', 'data': {}}, {}),
            (ListViewSetWithRequestSerializer, {}, {}),
    ),
)
def test_meta_data_serializer_success(mocker, viewset, meta_data, result):
    list_view = viewset(request=None, action='list', format_kwarg=None)
    mocker_meta_data = mocker.patch.object(list_view, 'get_meta_data')
    mocker_meta_data.return_value = meta_data

    assert list_view.get_meta_serializer_data() == result


@pytest.mark.parametrize(
    'meta_data,error_class',
    (
            ({'data': {}}, KeyError),
            ({'data': 'test'}, KeyError),
            ({'data': None}, KeyError),
            ({'data': []}, KeyError),
            ({}, KeyError),
            ([], AttributeError),
            ([{'char_field': 'test'}], AttributeError),
    ),
)
def test_meta_data_serializer_error(mocker, meta_data, error_class):
    list_view = ListViewSetWithMetaSerializer(request=None, action='list', format_kwarg=None)
    mocker_meta_data = mocker.patch.object(list_view, 'get_meta_data')
    mocker_meta_data.return_value = meta_data

    with pytest.raises(error_class):
        list_view.get_meta_serializer_data()
